import os
import time

import pandas as pd

from acrocord.utils import cache


def test_name():
    assert cache.name("a", "b", "c", 1) == "a_b_c_1"


def test_set_cache_folder_exists():
    cache.set_cache_folder()
    assert os.path.exists(cache.cache_folder)


def test_size_cache():
    cache.clean_cache()
    assert cache.folder_size_gio() == 0
    cache.set_cache_folder()
    cache.write(pd.DataFrame(index=range(1000), columns=range(1000)), "test")
    assert cache.folder_size_gio() > 0
    cache.clean_cache()
    assert cache.folder_size_gio() == 0


def test_write(
        get_connection,
        get_connection_cache,
        get_example_dataframe_building
):
    get_connection.create_schema("main")
    get_connection.write_table(get_example_dataframe_building, "test")
    get_connection_cache.create_schema("main")
    get_connection_cache.write_table(get_example_dataframe_building, "test")
    get_connection.read_table("main.test")
    get_connection_cache.read_table("main.test")

    start = time.time()
    df1 = get_connection.read_table("main.test")
    end = time.time()
    t1 = end - start

    start = time.time()
    df2 = get_connection_cache.read_table("main.test")
    end = time.time()

    t2 = end - start
    assert t2 < t1
    print(t1, t2)
    assert df1.equals(df2)



# Copyright 2023 Eurobios
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

import pytest

from acrocord.test import constraints


def test_example_span_table_constraint(
        get_example_dataframe_building):
    data = get_example_dataframe_building
    example_constraints = {
        "first_constraint": (constraints.unique, {"columns": ["building_id"]}),
        "second_constraint": (
            constraints.data_type,
            {'columns': ['building_id'], 'dtype': 'int64'}),
        "forth_constraint":
            (constraints.nb_unique_index,
             {"columns": ["building_id"], "minimum": 0,
              "maximum": 10}),
        "fifth_constraint": (constraints.quantile,
                             {"columns": ["building_id"], "q": 0.99,
                              "threshold": 1000}),
        "third_constraint": (
            constraints.not_nullable, {"columns": ["building_id"]})
    }

    dc = constraints.DataConstraints(data)
    dc.add_constraint(example_constraints)
    dc.test()


def test_example_span_table_constraint_raise(get_example_dataframe_building):
    data = get_example_dataframe_building
    example_constraints = {
        "first_constraint": (constraints.eligible_data_type, {}),
        "second_constraint": (
            constraints.data_type, {'columns': 'building_id'}),

    }

    dc = constraints.DataConstraints(data)
    dc.add_constraint(example_constraints)
    with pytest.raises(AssertionError):
        dc.test()



# Copyright 2023 Eurobios
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.#
def print_date(connection, schema: str, table: str):
    from acrocord.misc import execution
    with execution.silence_stdout():
        data = connection.read_table("public.informations_table").query(
            f"nom=='{table}'").query(f"schema=='{schema}'")
    if len(data) > 0:
        try:
            msg = "[info] creation date :"
            msg += "-" * (40 - len(msg)) + "  "
            msg += data["date_creation"].iloc[0][:-10]
            execution.print_(msg, "Grey")
            msg = "[info] author name :"
            msg += "-" * (40 - len(msg)) + "  "
            msg += data["auteur_creation"].iloc[0]
            execution.print_(msg, "Grey")
        except TypeError as error:
            execution.print_(error, color="Red")



# Copyright 2023 Eurobios
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

import asyncio
import pytest


def test_password_auth(get_connection):
    from acrocord.utils.connect import password_auth
    try:
        assert isinstance(password_auth(), str)
    except (PermissionError, FileNotFoundError):
        pass


def test_connect_literal():
    from acrocord.utils.connect import connect_psql_server
    str_ = "Engine(postgresql+psycopg2://user:***@localhost:5432/name)"

    assert str(connect_psql_server(
        username="user",
        connection=dict(
            password="pass",
            sslmode="require",
            host="localhost",
            port="5432",
            ssh=False, dbname="name"))) == str_


@pytest.mark.asyncio
async def test_connect_async_main(postgresql):
    from acrocord.utils.connect import connect_psql_server
    connect_psql_server(username=postgresql.info.user,
                        connection=dict(
                            password=" ",
                            sslmode="require",
                            host=postgresql.info.host,
                            port=postgresql.info.port,
                            ssh=False, dbname=postgresql.info.dbname), async_=True)



# Copyright 2023 Eurobios
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.#
import io
from typing import Iterator, Optional

import pandas as pd

from acrocord.misc import execution


@execution.execution_time
def insert(connection: "ConnectDatabase", data: pd.DataFrame, table_name: str,
           chunksize: int = 1000):
    storage = io.StringIO()
    data.to_csv(storage, sep='\t', header=False,
                index=False, encoding="utf8", chunksize=chunksize)
    storage.seek(0)
    sii = StringIteratorIO(storage)
    connection = connection.engine.connect()
    columns = tuple(f'{c}' for c in data.columns)

    with connection.connection.cursor() as cursor:
        cursor.execute(f'SET search_path TO {table_name.split(".")[0]}')
        cursor.copy_from(sii,
                         table_name.split(".")[1], null="",
                         columns=columns, sep="\t")
        cursor.connection.commit()
        cursor.close()


class StringIteratorIO(io.TextIOBase):
    def __init__(self, iter_: Iterator[str]):
        self._iter = iter_
        self._buff = ''

    def readable(self) -> bool:
        return True

    def _read1(self, n_line: Optional[int] = None) -> str:
        while not self._buff:
            try:
                self._buff = next(self._iter)
            except StopIteration:
                break
        ret = self._buff[:n_line]
        self._buff = self._buff[len(ret):]
        return ret

    def read(self, n: Optional[int] = None) -> str:
        line = []
        if n is None or n < 0:
            while True:
                line_str = self._read1()
                if not line_str:
                    break
                line.append(line_str)
        else:
            while n > 0:
                line_str = self._read1(n)
                if not line_str:
                    break
                n -= len(line_str)
                line.append(line_str)
        return ''.join(line)



# Copyright 2023 Eurobios
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.#
"""
This module aims  to specify data types that are consistent with
the usage of :object:`pypgsql`
"""
import warnings
from typing import List

import numpy as np
import pandas as pd


class EligibleDataType:

    # nullable integers
    INTEGER_64 = pd.Int64Dtype()  # [-9_223_372_036_854_775_808   to      9_223_372_036_854_775_807]
    INTEGER_32 = pd.Int32Dtype()  # [-2_147_483_648               to      2_147_483_647]
    INTEGER_16 = pd.Int16Dtype()  # [-32_768                      to      32_767]
    INTEGER_8 = pd.Int8Dtype()  # [-128                         to      127]
    INTEGER = INTEGER_64

    # float
    FLOAT_16 = np.float16
    FLOAT_32 = np.float32
    FLOAT_64 = np.float64
    FLOAT = FLOAT_64

    # nullable boolean
    BOOLEAN = pd.BooleanDtype()

    STRING = pd.StringDtype()

    # time
    DATE_TIME = np.dtype('datetime64[ns]')

    @classmethod
    def get_list(cls) -> List:
        """

        Returns
        -------
            list of eligible data types
        """
        return [
            cls.BOOLEAN,
            cls.INTEGER,
            cls.INTEGER_8,
            cls.INTEGER_16,
            cls.INTEGER_32,
            cls.INTEGER_64,
            cls.FLOAT_16,
            cls.FLOAT_32,
            cls.FLOAT_64,
            cls.FLOAT,
            cls.BOOLEAN,
            cls.STRING,
            cls.DATE_TIME
        ]


def warning_type(dataframe: pd.DataFrame, verbose: int):
    for column, column_type in dataframe.dtypes.items():
        if verbose == 2:
            if str(column_type) not in EligibleDataType.get_list():
                return warnings.warn(
                    "The types of this dataframe are not eligible types")
        if verbose > 2:
            if str(column_type) not in EligibleDataType.get_list():
                msg = f"The type of the column {column} " \
                      f"is {column_type} which is not an eligible type."
                warnings.warn(msg)



# Copyright 2023 Eurobios
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.#
import getpass
import sys
import tempfile


def pass_(path="."):
    """

    Parameters
    ----------
    path

    Returns
    -------

    """
    password = getpass.getpass(prompt="Enter the Password:")
    tmp = tempfile.NamedTemporaryFile(dir=path, delete=False)
    tmp.write(bytearray(password, encoding="utf-8"))


def pass_win():
    return getpass.win_getpass(prompt="Enter the Password:", stream=sys.stdout)


if __name__ == "__main__":
    pass_(sys.argv[1])



# Copyright 2023 Eurobios
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
import pytest
from sqlalchemy.exc import ProgrammingError

from acrocord.utils import auxiliaries


def test_execute_sql_cmd_raise_error(get_connection):
    sql_cmd_with_error = "select *"
    with pytest.raises(ProgrammingError) as e_info:
        auxiliaries.execute_sql_cmd(get_connection, sql_cmd_with_error)


def test_drop_table():
    assert auxiliaries.drop_table("test") == "DROP TABLE IF " \
                                             "EXISTS test CASCADE;"


def test_count():
    assert auxiliaries.count("test") == "SELECT COUNT(*) FROM test;"


def test_list_table():
    assert auxiliaries.list_table(
        "test") == f"SELECT * FROM information_schema.tables " \
                   f"WHERE table_schema = 'test'"


def test_data_types():
    a, b, c, d = "int32", "int4", "datetime64[s]", "timestamp"
    assert auxiliaries.db_data_type(a) == b
    assert auxiliaries.db_data_type(b, invert=True) == a
    assert auxiliaries.db_data_type(c) == d
    assert auxiliaries.db_data_type(d, invert=True) == c



# Copyright 2023 Eurobios
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

import warnings

import pandas as pd


def test_dataframe_equals_reading_writing(get_connection,
                                          get_example_data_frame):

    get_connection.write_table(get_example_data_frame, 'test.test_data')
    dataframe_read = get_connection.read_table('test.test_data')
    get_connection.drop_table('test.test_data')
    assert get_example_data_frame.equals(
        dataframe_read), "dataframe and dataframe_read are the same"


def test_warning_types(get_example_data_frame):
    from acrocord.utils.types import warning_type
    dataframe = get_example_data_frame.astype(int)
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        warning_type(dataframe, 3)
        assert len(w) == dataframe.shape[1]


def test_warning_types_64(get_example_data_frame):
    from acrocord.utils.types import warning_type
    dataframe = get_example_data_frame.astype("int64")
    warning_type(dataframe, 3)


def test_get_metadata(get_connection, get_example_data_frame):
    get_connection.write_table(get_example_data_frame, 'test.dataframe')
    df_meta = get_connection.get_metadata('test.dataframe')
    error_msg = "The number of line in meta data does correspond to the number of columns "
    number_columns = len(get_connection.get_columns('test.dataframe'))
    get_connection.drop_table('test.dataframe')
    assert len(df_meta) == number_columns, error_msg


def test_foreign_keys(get_connection, get_example_data_frame,
                      get_example_data_frame_other):
    df1 = get_example_data_frame
    df2 = get_example_data_frame_other
    get_connection.write_table(df1, "test.main", primary_key='a')
    get_connection.write_table(df2, "test.other",
                               foreign_keys={'a': "test.main"})


def test_get_shape(get_connection, get_example_data_frame):
    get_connection.write_table(get_example_data_frame, "test.main", primary_key='a')
    s = get_connection.get_shape("test.main")
    assert (s == get_example_data_frame.shape)


def test_query_where(get_connection, get_example_data_frame):
    get_connection.write_table(get_example_data_frame, "test.main",
                               primary_key='a')
    df = get_connection.read_table("test.main", where="a in (155)")
    assert len(df) == 1


def test_query_limit(get_connection, get_example_data_frame):
    get_connection.write_table(get_example_data_frame, "test.main",
                               primary_key='a')
    df = get_connection.read_table("test.main", limit=2)
    assert len(df) == 2


def test_create_insert(get_connection, get_example_data_frame):
    get_connection.write_table(get_example_data_frame, "test.main")
    get_connection.create_insert(get_example_data_frame, "test.main")
    df = get_connection.read_table("test.main")
    assert len(df) == len(get_example_data_frame) * 2


def test_create_insert_create(get_connection, get_example_data_frame):
    get_connection.drop_table("test.main")
    get_connection.create_insert(get_example_data_frame, "test.main")
    df = get_connection.read_table("test.main")
    assert len(df) == len(get_example_data_frame)


def test_create_drop_create_column(get_connection, get_example_data_frame):
    get_connection.write_table(get_example_data_frame, "test.main")
    get_connection.drop_column("test.main", "b")
    df = get_connection.read_table("test.main")
    assert df.shape[1] == get_example_data_frame.shape[1] - 1

    get_connection.add_columns(
        "test.main",
        get_example_data_frame[["a", "b"]],
        index="a"
    )


def test_copy(get_connection, get_example_data_frame_other):
    get_connection.write_table(get_example_data_frame_other, "test.main")
    get_connection.copy("test.main", "test.main_copy")
    df1 = get_connection.read_table("test.main")
    df2 = get_connection.read_table("test.main_copy")
    assert df1.equals(df2)


def test_get_table_names(get_connection, get_example_data_frame_other):
    get_connection.write_table(get_example_data_frame_other, "test.main")
    assert len(get_connection.get_table_names("test")) > 0


def test_get_view_names(get_connection, get_example_data_frame_other):
    get_connection.write_table(get_example_data_frame_other, "test.main")
    get_connection.execute("create view public.test as select * from test.main")
    assert len(get_connection.get_view_names("public")) > 0


def test_create_schema(get_connection, get_example_data_frame):
    get_connection.create_schema("test_ci")
    get_connection.write_table(get_example_data_frame, "test_ci.test_schema")
    assert "test_ci" in get_connection.get_schema_names()


# def test_geopandas(get_connection, get_example_data_frame):
#     import geopandas
#     connection = get_connection
#     dataframe = get_example_data_frame
#     gdf = geopandas.GeoDataFrame(
#         dataframe,
#         geometry=geopandas.points_from_xy(dataframe.lon, dataframe.lat))
#     connection.write_table(gdf, "test.test_lat_lon")
#
#


def test_list_table(get_connection):
    list_table = get_connection.list_table("public")
    print(list_table)


def test_update(get_connection, get_example_data_frame_other):
    get_connection.write_table(get_example_data_frame_other, "test.other")
    get_connection.update("test.other", where="a=155",
                          value="150", column="a")
    df_read = get_connection.read_table("test.other")
    assert 150 in df_read["a"].values


def test_get_dtypes(get_connection, get_example_data_frame_other):
    get_connection.write_table(get_example_data_frame_other, "test.other")
    res = get_connection.get_dtypes("test.other")
    assert isinstance(res, pd.DataFrame)
    assert len(res) == len(get_connection.get_columns("test.other"))


def test_drop(get_connection, get_example_data_frame, get_example_data_frame_other):
    get_connection.write_table(get_example_data_frame_other, "test.other")
    get_connection.write_table(get_example_data_frame, "test.main")
    get_connection.drop("test.main", "table", "a")
    assert "a" not in get_connection.get_columns("test.main")
    get_connection.write_table(get_example_data_frame_other, "test.main")


def test_execute(get_connection, get_example_data_frame):
    get_connection.write_table(get_example_data_frame, "test.main",
                               primary_key='a')
    get_connection.execute("select * from test.main")


def test_names_empty_list(get_connection):
    get_connection.execute("DROP SCHEMA if exists test_names")
    get_connection.create_schema("test_names")
    assert len(get_connection.get_table_names("test_names")) == 0
    get_connection.execute("DROP SCHEMA test_names")
